"""Real time OHLCV candles without middleware boxes.

This module is more or less indepedent from the rest of the codebase,
as it operates on raw JSON-RPC connection and raw DataFrame data structures.
"""

